/* $begin tinymain */
/*
* tiny.c - A simple, iterative HTTP/1.0 Web server that uses the 
*     GET method to serve static and dynamic content.
*/
#include "csapp.h"



void doit(int fd);
void read_requesthdrs(rio_t *rp, char* hostname, char *req_headers);
int parse_uri(char *uri,char*hostname, char *filename, char *cgiargs);
void serve_static(int fd, char *filename, int filesize);
void get_filetype(char *filename, char *filetype);
void serve_dynamic(int fd, char *filename, char *cgiargs);
void clienterror(int fd, char *cause, char *errnum, 
                char *shortmsg, char *longmsg);


static const char *user_agent_hdr = "User-Agent: Mozilla/5.0 (X11; Linux x86_64; rv:10.0.3) Gecko/20120305 Firefox/10.0.3\r\n";
static const char *accept_hdr = "Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8\r\n";
static const char *accept_encoding_hdr = "Accept-Encoding: gzip, deflate\r\n";


int main(int argc, char **argv) 
{
    int listenfd, connfd;
    char hostname[MAXLINE], port[MAXLINE];
    socklen_t clientlen;
    struct sockaddr_storage clientaddr;

/* Check command line args */
    if (argc != 2) {
       fprintf(stderr, "usage: %s <port>\n", argv[0]);
       exit(1);
   }

   listenfd = Open_listenfd(argv[1]);
   while (1) {
        clientlen = sizeof(clientaddr);
        connfd = Accept(listenfd, (SA *)&clientaddr, &clientlen); //line:netp:tiny:accept
        Getnameinfo((SA *) &clientaddr, clientlen, hostname, MAXLINE, 
                    port, MAXLINE, 0);
        printf("Accepted connection from (%s, %s)\n", hostname, port);
        doit(connfd);                                             //line:netp:tiny:doit
        Close(connfd);                                            //line:netp:tiny:close
    }
}
/* $end tinymain */

/*
* doit - handle one HTTP request/response transaction
*/
/* $begin doit */
void doit(int fd) 
{
    int is_static;
    char buf[MAXLINE], method[MAXLINE], uri[MAXLINE], version[MAXLINE];
    char filename[MAXLINE], cgiargs[MAXLINE], hostname[MAXLINE];
    char req_headers[MAXLINE], req_line[MAXLINE], req_overall[MAXLINE];
    rio_t rio_client, rio_server;

    int proxy_server_fd; // proxy serves as a client
    

    req_line[0] = req_headers[0] = req_overall[0] = '\0';
    hostname[0] = '\0';
/* Read request line and headers */
    Rio_readinitb(&rio_client, fd);
    if (!Rio_readlineb(&rio_client, buf, MAXLINE))  //line:netp:doit:readrequest
        return;
    printf("%s", buf);
    sscanf(buf, "%s %s %s", method, uri, version);       //line:netp:doit:parserequest
    if (strcasecmp(method, "GET")) {                     //line:netp:doit:beginrequesterr
        clienterror(fd, method, "501", "Not Implemented",
            "Tiny does not implement this method");
        return;
    }                                                    //line:netp:doit:endrequesterr
    

    /* Parse URI from GET request */
    is_static = parse_uri(uri, hostname, filename, cgiargs);       //line:netp:doit:staticcheck
    
    sprintf(req_line, "%sGET %s HTTP/1.0\r\n", req_line, filename);     // only HTTP/1.0 and GET method are supported
    read_requesthdrs(&rio_client, hostname, req_headers);                //line:netp:doit:readrequesthdrs
    strcpy(req_overall, req_line);
    strcat(req_overall, req_headers);
    
    // proxy connects with server
    printf("Try clientfd..hostname = %s \n", hostname);

    proxy_server_fd = Open_clientfd(hostname, "80");
    
    Rio_readinitb(&rio_server, proxy_server_fd);
    
    Rio_writen(proxy_server_fd, req_overall, strlen(req_overall));
    int n;
    while ((n = Rio_readlineb(&rio_server, buf, MAXLINE)) != 0) {
        printf("%s", buf);
        Rio_writen(fd, buf, n);
    }
    
    Close(proxy_server_fd);

}
/* $end doit */

/*
 * Function: read_requesthdrs - read HTTP request headers
 * If specified in headers, then we set Host header according to browser's original host values.
 * By default, the Host value is parsed from GET request line
 * 
 */
void read_requesthdrs(rio_t *rp, char *host, char *headers) 
{  
    char buf[MAXLINE];
    char additionals[MAXLINE];
    headers[0] = additionals[0] = '\0';
    
    while (1) {          //line:netp:readhdrs:checkterm
        Rio_readlineb(rp, buf, MAXLINE);
        if (strcmp(buf, "\r\n") == 0) break;
        //printf("%s", buf);
        if (strstr(buf, "Host: ")) {
            strcpy(host, buf + strlen("Host: "));
            continue;
        }
        if (strstr(buf, "User-Agent: ") == buf || strstr(buf, "Accept: ") == buf
            || strstr(buf, "Accept: ") == buf || strstr(buf, "Accept-Encoding: ") == buf
            || strstr(buf, "Connection: ") == buf || strstr(buf, "Proxy-Connection: ") == buf)
            continue;
        sprintf(additionals, "%s%s", additionals, buf);
        
    }

    sprintf(headers, "%sHost: %s\r\n", headers, host);
    sprintf(headers, "%s%s", headers, user_agent_hdr);
    sprintf(headers, "%s%s", headers, accept_hdr);
    sprintf(headers, "%s%s", headers, accept_encoding_hdr);
    sprintf(headers, "%sConnection: close\r\n", headers);
    sprintf(headers, "%sProxy-Connection: close\r\n", headers);
    sprintf(headers, "%s%s\r\n", headers, additionals);
    
    
}
/* $end read_requesthdrs */

/*
* parse_uri - parse URI into filename and CGI args
*             return 0 if dynamic content, 1 if static
*/
/* $begin parse_uri */
int parse_uri(char *uri, char *hostname, char *filename, char *cgiargs) 
{
    
    
    /* Simple host extraction */
    char *hostStart, *hostEnd;

    if (strstr(uri, "http://")) 
        hostStart = uri + strlen("http://");
    else
        hostStart = uri;
    if ((hostEnd = strstr(hostStart, "/")) == NULL)
            hostEnd = uri + strlen(uri);
    strncpy(hostname, hostStart, hostEnd - hostStart);
    hostname[(int)(hostEnd - hostStart)] = '\0';
    if (strlen(hostname) == 0)
        strcpy(hostname, "localhost");
    //strcat(hostname, "\r\n");

    if (strlen(hostEnd) == 0)
        strcpy(filename, "/"); 
    else
        strcpy(filename, hostEnd);
    return 0;
}
/* $end parse_uri */

/*
* serve_static - copy a file back to the client 
*/
/* $begin serve_static */
void serve_static(int fd, char *filename, int filesize) 
{
    int srcfd;
    char *srcp, filetype[MAXLINE], buf[MAXBUF];

    /* Send response headers to client */
    get_filetype(filename, filetype);       //line:netp:servestatic:getfiletype
    sprintf(buf, "HTTP/1.0 200 OK\r\n");    //line:netp:servestatic:beginserve
    sprintf(buf, "%sServer: Tiny Web Server\r\n", buf);
    sprintf(buf, "%sConnection: close\r\n", buf);
    sprintf(buf, "%sContent-length: %d\r\n", buf, filesize);
    sprintf(buf, "%sContent-type: %s\r\n\r\n", buf, filetype);
    Rio_writen(fd, buf, strlen(buf));       //line:netp:servestatic:endserve
    printf("Response headers:\n");
    printf("%s", buf);

    /* Send response body to client */
    srcfd = Open(filename, O_RDONLY, 0);    //line:netp:servestatic:open
    srcp = Mmap(0, filesize, PROT_READ, MAP_PRIVATE, srcfd, 0);//line:netp:servestatic:mmap
    Close(srcfd);                           //line:netp:servestatic:close
    Rio_writen(fd, srcp, filesize);         //line:netp:servestatic:write
    Munmap(srcp, filesize);                 //line:netp:servestatic:munmap
}

/*
* get_filetype - derive file type from file name
*/
void get_filetype(char *filename, char *filetype) 
{
    if (strstr(filename, ".html"))
        strcpy(filetype, "text/html");
    else if (strstr(filename, ".gif"))
        strcpy(filetype, "image/gif");
    else if (strstr(filename, ".png"))
        strcpy(filetype, "image/png");
    else if (strstr(filename, ".jpg"))
        strcpy(filetype, "image/jpeg");
    else
        strcpy(filetype, "text/plain");
}  
/* $end serve_static */

/*
* serve_dynamic - run a CGI program on behalf of the client
*/
/* $begin serve_dynamic */
void serve_dynamic(int fd, char *filename, char *cgiargs) 
{
    char buf[MAXLINE], *emptylist[] = { NULL };

/* Return first part of HTTP response */
    sprintf(buf, "HTTP/1.0 200 OK\r\n"); 
    Rio_writen(fd, buf, strlen(buf));
    sprintf(buf, "Server: Tiny Web Server\r\n");
    Rio_writen(fd, buf, strlen(buf));

    if (Fork() == 0) { /* Child */ //line:netp:servedynamic:fork
        /* Real server would set all CGI vars here */
        setenv("QUERY_STRING", cgiargs, 1); //line:netp:servedynamic:setenv
        Dup2(fd, STDOUT_FILENO);         /* Redirect stdout to client */ //line:netp:servedynamic:dup2
        Execve(filename, emptylist, environ); /* Run CGI program */ //line:netp:servedynamic:execve
    }
    Wait(NULL); /* Parent waits for and reaps child */ //line:netp:servedynamic:wait
}
/* $end serve_dynamic */

/*
* clienterror - returns an error message to the client
*/
/* $begin clienterror */
void clienterror(int fd, char *cause, char *errnum, 
 char *shortmsg, char *longmsg) 
{
    char buf[MAXLINE], body[MAXBUF];

/* Build the HTTP response body */
    sprintf(body, "<html><title>Tiny Error</title>");
    sprintf(body, "%s<body bgcolor=""ffffff"">\r\n", body);
    sprintf(body, "%s%s: %s\r\n", body, errnum, shortmsg);
    sprintf(body, "%s<p>%s: %s\r\n", body, longmsg, cause);
    sprintf(body, "%s<hr><em>The Tiny Web server</em>\r\n", body);

/* Print the HTTP response */
    sprintf(buf, "HTTP/1.0 %s %s\r\n", errnum, shortmsg);
    Rio_writen(fd, buf, strlen(buf));
    sprintf(buf, "Content-type: text/html\r\n");
    Rio_writen(fd, buf, strlen(buf));
    sprintf(buf, "Content-length: %d\r\n\r\n", (int)strlen(body));
    Rio_writen(fd, buf, strlen(buf));
    Rio_writen(fd, body, strlen(body));
}
/* $end clienterror */
